@@ -78,6 +78,7 @@ module Language.Haskell.TH.Datatype
   -- * Type variable manipulation
   , TypeSubstitution(..)
   , quantifyType
+  , freeVariablesWellScoped
   , freshenFreeVariables
 
   -- * 'Pred' functions
@@ -100,6 +101,7 @@ module Language.Haskell.TH.Datatype
 
   -- * Type simplification
   , resolveTypeSynonyms
+  , resolveKindSynonyms
   , resolvePredSynonyms
   , resolveInfixT
 
@@ -121,6 +123,8 @@ import           Data.List (nub, find, union, (\\))
 import           Data.Map (Map)
 import qualified Data.Map as Map
 import           Data.Maybe
+import qualified Data.Set as Set
+import           Data.Set (Set)
 import qualified Data.Traversable as T
 import           Control.Monad
 import           Language.Haskell.TH
@@ -136,6 +140,7 @@ import           GHC.Generics (Generic)
 
 #if !MIN_VERSION_base(4,8,0)
 import           Control.Applicative (Applicative(..), (<$>))
+import           Data.Monoid (Monoid(..))
 #endif
 
 -- | Normalized information about newtypes and data types.
@@ -1085,20 +1090,61 @@ mergeArgumentKinds _ _ = (Map.empty, [])
 #endif
 
 -- | Expand all of the type synonyms in a type.
+--
+-- Note that this function will drop parentheses as a side effect.
 resolveTypeSynonyms :: Type -> Q Type
 resolveTypeSynonyms t =
   let f :| xs = decomposeType t
 
-      notTypeSynCase = foldl AppT f <$> mapM resolveTypeSynonyms xs in
+      notTypeSynCase :: Type -> Q Type
+      notTypeSynCase ty = foldl AppT ty <$> mapM resolveTypeSynonyms xs
 
-  case f of
-    ConT n ->
-      do mbInfo <- reifyMaybe n
-         case mbInfo of
-           Just (TyConI (TySynD _ synvars def))
-             -> resolveTypeSynonyms $ expandSynonymRHS synvars xs def
-           _ -> notTypeSynCase
-    _ -> notTypeSynCase
+      expandCon :: Name -- The Name to check whether it is a type synonym or not
+                -> Type -- The argument type to fall back on if the supplied
+                        -- Name isn't a type synonym
+                -> Q Type
+      expandCon n ty = do
+        mbInfo <- reifyMaybe n
+        case mbInfo of
+          Just (TyConI (TySynD _ synvars def))
+            -> resolveTypeSynonyms $ expandSynonymRHS synvars xs def
+          _ -> notTypeSynCase ty
+
+  in case f of
+       ForallT tvbs ctxt body ->
+         ForallT `fmap` mapM resolve_tvb_syns tvbs
+                   `ap` mapM resolvePredSynonyms ctxt
+                   `ap` resolveTypeSynonyms body
+       SigT ty ki -> do
+         ty' <- resolveTypeSynonyms ty
+         ki' <- resolveKindSynonyms ki
+         notTypeSynCase $ SigT ty' ki'
+       ConT n -> expandCon n (ConT n)
+#if MIN_VERSION_template_haskell(2,11,0)
+       InfixT t1 n t2 -> do
+         t1' <- resolveTypeSynonyms t1
+         t2' <- resolveTypeSynonyms t2
+         expandCon n (InfixT t1' n t2')
+       UInfixT t1 n t2 -> do
+         t1' <- resolveTypeSynonyms t1
+         t2' <- resolveTypeSynonyms t2
+         expandCon n (UInfixT t1' n t2')
+#endif
+       _ -> notTypeSynCase f
+
+-- | Expand all of the type synonyms in a 'Kind'.
+resolveKindSynonyms :: Kind -> Q Kind
+#if MIN_VERSION_template_haskell(2,8,0)
+resolveKindSynonyms = resolveTypeSynonyms
+#else
+resolveKindSynonyms = return -- One simply couldn't put type synonyms into
+                             -- kinds on old versions of GHC.
+#endif
+
+-- | Expand all of the type synonyms in a the kind of a 'TyVarBndr'.
+resolve_tvb_syns :: TyVarBndr -> Q TyVarBndr
+resolve_tvb_syns tvb@PlainTV{}  = return tvb
+resolve_tvb_syns (KindedTV n k) = KindedTV n <$> resolveKindSynonyms k
 
 expandSynonymRHS ::
   [TyVarBndr] {- ^ Substitute these variables... -} ->
@@ -1156,8 +1202,11 @@ typeToPred t =
 decomposeType :: Type -> NonEmpty Type
 decomposeType = go []
   where
-    go args (AppT f x) = go (x:args) f
-    go args t          = t :| args
+    go args (AppT f x)  = go (x:args) f
+#if MIN_VERSION_template_haskell(2,11,0)
+    go args (ParensT t) = go args t
+#endif
+    go args t           = t :| args
 
 -- 'NonEmpty' didn't move into base until recently. Reimplementing it locally
 -- saves dependencies for supporting older GHCs
@@ -1294,11 +1343,151 @@ conHasRecord recName info =
 -- contrast with being dependent upon the Ord instance for 'Name')
 quantifyType :: Type -> Type
 quantifyType t
-  | null vs   = t
-  | otherwise = ForallT (PlainTV <$> vs) [] t
+  | null tvbs
+  = t
+  | ForallT tvbs' ctxt' t' <- t -- Collapse two consecutive foralls (#63)
+  = ForallT (tvbs ++ tvbs') ctxt' t'
+  | otherwise
+  = ForallT tvbs [] t
   where
-    vs = freeVariables t
+    tvbs = freeVariablesWellScoped [t]
 
+-- | Take a list of 'Type's, find their free variables, and sort them
+-- according to dependency order.
+--
+-- As an example of how this function works, consider the following type:
+--
+-- @
+-- Proxy (a :: k)
+-- @
+--
+-- Calling 'freeVariables' on this type would yield @[a, k]@, since that is
+-- the order in which those variables appear in a left-to-right fashion. But
+-- this order does not preserve the fact that @k@ is the kind of @a@. Moreover,
+-- if you tried writing the type @forall a k. Proxy (a :: k)@, GHC would reject
+-- this, since GHC would demand that @k@ come before @a@.
+--
+-- 'freeVariablesWellScoped' orders the free variables of a type in a way that
+-- preserves this dependency ordering. If one were to call
+-- 'freeVariablesWellScoped' on the type above, it would return
+-- @[k, (a :: k)]@. (This is why 'freeVariablesWellScoped' returns a list of
+-- 'TyVarBndr's instead of 'Name's, since it must make it explicit that @k@
+-- is the kind of @a@.)
+--
+-- 'freeVariablesWellScoped' guarantees the free variables returned will be
+-- ordered such that:
+--
+-- 1. Whenever an explicit kind signature of the form @(A :: K)@ is
+--    encountered, the free variables of @K@ will always appear to the left of
+--    the free variables of @A@ in the returned result.
+--
+-- 2. The constraint in (1) notwithstanding, free variables will appear in
+--    left-to-right order of their original appearance.
+--
+-- On older GHCs, this takes measures to avoid returning explicitly bound
+-- kind variables, which was not possible before @TypeInType@.
+freeVariablesWellScoped :: [Type] -> [TyVarBndr]
+freeVariablesWellScoped tys =
+  let fvs :: [Name]
+      fvs = freeVariables tys
+
+      varKindSigs :: Map Name Kind
+      varKindSigs = foldMap go_ty tys
+        where
+          go_ty :: Type -> Map Name Kind
+          go_ty (ForallT tvbs ctxt t) =
+            foldr (\tvb -> Map.delete (tvName tvb))
+                  (foldMap go_pred ctxt `mappend` go_ty t) tvbs
+          go_ty (AppT t1 t2) = go_ty t1 `mappend` go_ty t2
+          go_ty (SigT t k) =
+            let kSigs =
+#if MIN_VERSION_template_haskell(2,8,0)
+                  go_ty k
+#else
+                  mempty
+#endif
+            in case t of
+                 VarT n -> Map.insert n k kSigs
+                 _      -> go_ty t `mappend` kSigs
+          go_ty _ = mempty
+
+          go_pred :: Pred -> Map Name Kind
+#if MIN_VERSION_template_haskell(2,10,0)
+          go_pred = go_ty
+#else
+          go_pred (ClassP _ ts)  = foldMap go_ty ts
+          go_pred (EqualP t1 t2) = go_ty t1 `mappend` go_ty t2
+#endif
+
+      -- | Do a topological sort on a list of tyvars,
+      --   so that binders occur before occurrences
+      -- E.g. given  [ a::k, k::*, b::k ]
+      -- it'll return a well-scoped list [ k::*, a::k, b::k ]
+      --
+      -- This is a deterministic sorting operation
+      -- (that is, doesn't depend on Uniques).
+      --
+      -- It is also meant to be stable: that is, variables should not
+      -- be reordered unnecessarily.
+      scopedSort :: [Name] -> [Name]
+      scopedSort = go [] []
+
+      go :: [Name]     -- already sorted, in reverse order
+         -> [Set Name] -- each set contains all the variables which must be placed
+                       -- before the tv corresponding to the set; they are accumulations
+                       -- of the fvs in the sorted tvs' kinds
+
+                       -- This list is in 1-to-1 correspondence with the sorted tyvars
+                       -- INVARIANT:
+                       --   all (\tl -> all (`isSubsetOf` head tl) (tail tl)) (tails fv_list)
+                       -- That is, each set in the list is a superset of all later sets.
+         -> [Name]     -- yet to be sorted
+         -> [Name]
+      go acc _fv_list [] = reverse acc
+      go acc  fv_list (tv:tvs)
+        = go acc' fv_list' tvs
+        where
+          (acc', fv_list') = insert tv acc fv_list
+
+      insert :: Name       -- var to insert
+             -> [Name]     -- sorted list, in reverse order
+             -> [Set Name] -- list of fvs, as above
+             -> ([Name], [Set Name])   -- augmented lists
+      insert tv []     []         = ([tv], [kindFVSet tv])
+      insert tv (a:as) (fvs:fvss)
+        | tv `Set.member` fvs
+        , (as', fvss') <- insert tv as fvss
+        = (a:as', fvs `Set.union` fv_tv : fvss')
+
+        | otherwise
+        = (tv:a:as, fvs `Set.union` fv_tv : fvs : fvss)
+        where
+          fv_tv = kindFVSet tv
+
+         -- lists not in correspondence
+      insert _ _ _ = error "scopedSort"
+
+      kindFVSet n =
+        maybe Set.empty (Set.fromList . freeVariables) (Map.lookup n varKindSigs)
+      ascribeWithKind n =
+        maybe (PlainTV n) (KindedTV n) (Map.lookup n varKindSigs)
+
+      -- An annoying wrinkle: GHCs before 8.0 don't support explicitly
+      -- quantifying kinds, so something like @forall k (a :: k)@ would be
+      -- rejected. To work around this, we filter out any binders whose names
+      -- also appear in a kind on old GHCs.
+      isKindBinderOnOldGHCs
+#if __GLASGOW_HASKELL__ >= 800
+        = const False
+#else
+        = (`elem` kindVars)
+          where
+            kindVars = freeVariables $ Map.elems varKindSigs
+#endif
+
+  in map ascribeWithKind $
+     filter (not . isKindBinderOnOldGHCs) $
+     scopedSort fvs
 
 -- | Substitute all of the free variables in a type with fresh ones
 freshenFreeVariables :: Type -> Q Type
