@@ -78,12 +78,16 @@ main =
 #endif
 #if MIN_VERSION_template_haskell(2,8,0)
      kindSubstTest
+     t59Test
+     t61Test
 #endif
 #if __GLASGOW_HASKELL__ >= 800
      t37Test
      polyKindedExTyvarTest
 #endif
      regressionTest44
+     t63Test
+     t70Test
 
 adt1Test :: IO ()
 adt1Test =
@@ -661,6 +665,58 @@ kindSubstTest =
        checkFreeVars ty      [k1]
        checkFreeVars substTy [k2]
        [| return () |])
+
+t59Test :: IO ()
+t59Test =
+  $(do k <- newName "k"
+       a <- newName "a"
+       let proxyAK  = ConT (mkName "Proxy") `AppT` SigT (VarT a) (VarT k)
+                        -- Proxy (a :: k)
+           expected = ForallT
+#if __GLASGOW_HASKELL__ >= 800
+                        [PlainTV k, KindedTV a (VarT k)]
+#else
+                        [KindedTV a (VarT k)]
+#endif
+                        [] proxyAK
+           actual = quantifyType proxyAK
+       unless (expected == actual) $
+         fail $ "quantifyType does not respect dependency order: "
+             ++ unlines [ "Expected: " ++ pprint expected
+                        , "Actual:   " ++ pprint actual
+                        ]
+       [| return () |])
+
+t61Test :: IO ()
+t61Test =
+  $(do let test :: Type -> Type -> Q ()
+           test orig expected = do
+             actual <- resolveTypeSynonyms orig
+             unless (expected == actual) $
+               fail $ "Type synonym expansion failed: "
+                   ++ unlines [ "Expected: " ++ pprint expected
+                              , "Actual:   " ++ pprint actual
+                              ]
+
+           idAppT = (ConT ''Id `AppT`)
+           a = mkName "a"
+       test (SigT (idAppT $ ConT ''Int) (idAppT StarT))
+            (SigT (ConT ''Int) StarT)
+#if MIN_VERSION_template_haskell(2,10,0)
+       test (ForallT [KindedTV a (idAppT StarT)]
+                     [idAppT (ConT ''Show `AppT` VarT a)]
+                     (idAppT $ VarT a))
+            (ForallT [KindedTV a StarT]
+                     [ConT ''Show `AppT` VarT a]
+                     (VarT a))
+#endif
+#if MIN_VERSION_template_haskell(2,11,0)
+       test (InfixT (idAppT $ ConT ''Int) ''Either (idAppT $ ConT ''Int))
+            (InfixT (ConT ''Int) ''Either (ConT ''Int))
+       test (ParensT (idAppT $ ConT ''Int))
+            (ConT ''Int)
+#endif
+       [| return () |])
 #endif
 
 #if __GLASGOW_HASKELL__ >= 800
@@ -760,3 +816,44 @@ regressionTest44 =
        unified  <- unifyTypes [intToInt, intToInt]
        unless (Map.null unified) (fail "regression test for ticket #44 failed")
        [| return () |])
+
+t63Test :: IO ()
+t63Test =
+  $(do a <- newName "a"
+       b <- newName "b"
+       t <- newName "T"
+       let tauType = ArrowT `AppT` VarT a `AppT` (ArrowT `AppT` VarT b
+                       `AppT` (ConT t `AppT` VarT a))
+           sigmaType = ForallT [PlainTV b] [] tauType
+           expected = ForallT [PlainTV a, PlainTV b] [] tauType
+           actual   = quantifyType sigmaType
+       unless (expected == actual) $
+         fail $ "quantifyType does not collapse consecutive foralls: "
+             ++ unlines [ "Expected: " ++ pprint expected
+                        , "Actual:   " ++ pprint actual
+                        ]
+       [| return () |])
+
+t70Test :: IO ()
+t70Test =
+  $(do a <- newName "a"
+       b <- newName "b"
+       let [aVar, bVar] = map VarT    [a, b]
+           [aTvb, bTvb] = map PlainTV [a, b]
+       let fvsABExpected = [aTvb, bTvb]
+           fvsABActual   = freeVariablesWellScoped [aVar, bVar]
+
+           fvsBAExpected = [bTvb, aTvb]
+           fvsBAActual   = freeVariablesWellScoped [bVar, aVar]
+
+           check expected actual =
+             unless (expected == actual) $
+               fail $ "freeVariablesWellScoped does not preserve left-to-right order: "
+                   ++ unlines [ "Expected: " ++ pprint expected
+                              , "Actual:   " ++ pprint actual
+                              ]
+
+       check fvsABExpected fvsABActual
+       check fvsBAExpected fvsBAActual
+
+       [| return () |])
